/*
 * Copyright (c) 1996-1999 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */





/************************************************************************
 * msg.c: message functions for libthread.
************************************************************************/





/************************************************************************
 * header files.
************************************************************************/

#include "threadP.h"






/************************************************************************
 * constants.
************************************************************************/

#define BUFSIZE 1024







/************************************************************************
 * prototypes for local functions.
************************************************************************/

static	void	msg_copy(void* to_buf, unsigned* to_sizep, void* from_buf, unsigned from_size);
static	void	queue_msg(tag_t tag, void* buf, unsigned size, Thread t);
static	ThrMessage*	find_match(thread_t tid, tag_t tag, unsigned do_poll);
static	bool	got_sigmsg(tag_t* tagp, unsigned* nitemsp, int* senderp, unsigned do_poll);
static	bool	got_qmsg(thread_t* tidp, tag_t* tagp, void* buf, unsigned* sizep, unsigned do_poll);
static	bool	got_filemsg(thread_t* tidp, tag_t* tagp, void* buf, unsigned* sizep, unsigned do_poll);
static	bool	got_socketmsg(thread_t* tidp, tag_t* tagp, void* buf, unsigned* sizep, bool do_poll);
static	int		chk_msg_pref(const char* func, thread_t* tidp, tag_t* tagp,
							 void* buf, unsigned* nitemsp,
							 bool do_poll, bool do_block, unsigned io_first);
static	void	add_all_bytype( WaitSet& wset, item_t itype );
static	void	add_all( WaitSet& wset );




/************************************************************************
 * inlined utility functions
************************************************************************/

inline
bool
msg_match( thread_t tid, tag_t tag,
		   thread_t desired_tid, tag_t desired_tag )
{
	// check tids for a match
	// (tids match if the desired tid is unspecified or
	// if the given tid is the desired tid)
	if( is_unspec_tid(desired_tid) || (tid == desired_tid) )
	{
		// the tids match - check tags for a match
		// we consider the tags to match if:
		//   the desired tag is unspecified, or
		//   the given tag exactly matches the desired tag, or
		//   the desired tag indicates any thread message and the
		//     given message comes from a thread
		if( (desired_tag == MSG_TAG_ANY) ||
			(desired_tag == tag) ||
			((desired_tag == MSG_TAG_THREAD) && (tag >= MSG_TAG_USER)) )
		{
			// the tags match too
			return true;
		}
	}

	return false;
}





/************************************************************************
 * interface functions
************************************************************************/


/************************************************************************
 * int msg_send(thread_t tid, tag_t tag, void* buf, unsigned size)
 *
 * send a message `buf' of size `size' and type 'tag' to thread 'tid'. 
 *
 * it is an error to specify an tid that does not belong to a thread,
 * or an tid that belongs to an invalid or exited thread as the 
 * destination of a message.
 *
 * if `tid' corresponds to a file descriptor (which may or may not have
 * been mapped for reading by some thread), a write is done on this file
 *
 * under Windows NT:
 *
 *    if `tid' denotes a SOCKET, a write is done on the SOCKET
 *
 * for thread receivers:
 *
 * if the receiver is blocked waiting for a message, the sender's buffers
 * are directly copied into the receiver's buffers (and message sizes are
 * updated). the receiver is unblocked. if the receiver is not ready
 * for the message, a buffer is allocated and the message copied into
 * it and queued at the receiver's message queue.
 *
 * if the receiver excepts a message smaller than what the sender is
 * transmitting, a warning message is printed, and an incomplete message
 * is copied. the sender may not be able to do anything about this, esp.
 * if the message was queued and the send returned successfully.
 *
 * the message queue for the recipient may be initialized here, instead
 * of during thread create.
************************************************************************/

int
msg_send(thread_t tid, tag_t tag, void* buf, unsigned size)
{
	unsigned 	nbytes;
    Thread		t;
    int			ret;


    THR_ENTER_LIB();

    if (!thr_init_done) {
        thr_init();
    }
    assert(thr_curr);


	// handle IO bindable destinations
	if( thr_tab->GetItem( tid )->IsIOBindable() )
	{
		int err = thr_tab->GetIOBindable(tid)->Write( buf, size, &nbytes );
		if( err == THR_OKAY )
		{
			ret = (int)nbytes;
		}
		thr_err( err );
		VRETURN;
	}

	// handle thread destinations
    if (!thr_tab->IsThread(tid)) {
        thr_report_error("bad tid %u\n", tid);
        RETURN(THR_ESRCH);
    }

    if (!is_thr_tag(tag)) {
        thr_report_error("invalid message tag %u\n", tag);
        RETURN(THR_EBADTAG);
    }

    t = thr_tab->GetThread(tid);
    if (t->flags.exited) {
        thr_report_error("sending msg to dead t%u\n", tid);
        RETURN(THR_ESRCH);
    }

    if ((t->flags.blocked) &&
		msg_match( thr_curr->id.tid, tag, *(t->msg.tidp), *(t->msg.tagp) ) ) {

        if (t->flags.polling) {
            queue_msg(tag, buf, size, t);
			ret = (int)size;
        }
        else {
            msg_copy(t->msg.buf, t->msg.nitemsp, buf, size);
			ret = (int)( (t->msg.nitemsp != NULL) ? *(t->msg.nitemsp) : 0 );
        }

		*(t->msg.tidp)			= thr_curr->id.tid;
		*(t->msg.tagp)			= tag;
        t->flags.blocked		= 0;
        t->flags.polling		= 0;

        ENQUEUE(thr_ready_q, t);
        DO_TRACE_0(("msg_send(%u,%u,%p,%u): unblocked receiver",
            tid, tag, buf, size));
    }
    else {
        queue_msg(tag, buf, size, t);
        DO_TRACE_0(("msg_send(%u,%u,%p,%u): queued message with receiver",
            tid, tag, buf, size));
		ret = (int)size;
    }

    CHECK_RESCHEDULE();
    thr_err(THR_OKAY);

finish:
    THR_LEAVE_LIB();
    return ret;
}





/************************************************************************
 * int msg_recv(thread_t* tidp, tag_t* tagp, void* buf, unsigned* nitemsp)
 *
 * wait for a message. if one is not available in the message queue
 * then block and invoke the scheduler. when a message arrives, the
 * sender copies it into the buffer, sets `*nitemsp' and unblocks the
 * receiver.
 *
 * only messages matching '*tidp' and '*tagp' are chosen.
 *
 * if messages are waiting in the local queue, the first one is
 * copied by the receiver and the sender is unblocked.
************************************************************************/

int
msg_recv(thread_t* tidp, tag_t* tagp, void* buf, unsigned* nitemsp)
{
    int  ret;
    char func[BUFSIZE+1];

    THR_ENTER_LIB();

    assert(tagp);
	assert(tidp);
    sprintf(func, "msg_recv(%u,%u,%p,%p)", *tidp, *tagp, buf, nitemsp);
    ret = chk_msg_pref(func, tidp, tagp, buf, nitemsp, false, false, 0);

    THR_LEAVE_LIB();
    return ret;
}





/************************************************************************
 * int msg_poll(tag_t* tagp, unsigned block)
 *
 * similar in function to `msg_recv', but only checks for messages.
 * no state change is made. if the boolean flag `block' is not set,
 * the function does not even block waiting for messages.
 *
 * if `block' is not set and no messages are found, the function returns
 * `-1' and sets the internal error to `THR_ENOMSG'.
************************************************************************/

int
msg_poll(thread_t* tidp, tag_t* tagp, unsigned block)
{
    int  ret;
    char func[BUFSIZE+1];

    THR_ENTER_LIB();

	assert(tidp);
    assert(tagp);
    sprintf(func, "msg_poll(%u,%u,%u)", *tidp, *tagp, block);
    ret = chk_msg_pref(func, tidp, tagp, 0, 0, true, block, 0);

    THR_LEAVE_LIB();
    return ret;
}



/************************************************************************
 * int msg_poll_preference(tag_t* tagp, unsigned block, unsigned io_first)
 *
 * similar in function to `msg_recv', but only checks for messages.
 * no state change is made. if the boolean flag `block' is not set,
 * the function does not even block waiting for messages.  If `io_first'
 * has a positive value then a check for a message from a file or socket
 * is done before checking for a msg from another thread.
 *
 * if `block' is not set and no messages are found, the function returns
 * `-1' and sets the internal error to `THR_ENOMSG'.
************************************************************************/

int
msg_poll_preference(thread_t* tidp, tag_t* tagp, unsigned block, unsigned io_first)
{
    int  ret;
    char func[BUFSIZE+1];

    THR_ENTER_LIB();

	assert(tidp);
    assert(tagp);
    sprintf(func, "msg_poll_preference(%u,%u,%u,%u)", *tidp, *tagp, block, io_first);
    ret = chk_msg_pref(func, tidp, tagp, 0, 0, true, block, io_first);

    THR_LEAVE_LIB();
    return ret;
}






/************************************************************************
 * int msg_bind(PDDESC fd, unsigned special,
 *              int (*will_block)(void *), void* desc, thread_t* ptid)
 *
 * bind file descriptor `fd' to the calling thread so that all _input_
 * coming thru this file is available as a message to the thread.
 *
 * the `special' flag indicates how `fd' should be treated. if not set,
 * bytes appearing on this `fd' will be copied to the receiver's buffers.
 * if `special' is set, the receiver will get an indication of the `fd'
 * but no bytes will be dequeued.
 *
 * If a non-null will_block function pointer is given, libthread assumes
 * that the data on the stream `fd' can be buffered, and uses 
 * `will_block(desc)' to determine if no buffered data is
 * present for the descriptor in question.
 *
 * The thread_t assigned to the given file descriptor is returned
 * in *ptid.  It is an error to specify NULL for ptid.
************************************************************************/

int
msg_bind(PDDESC fd,
		unsigned special,
		int (*will_block)(void *),
		void* desc,
		thread_t* ptid)
{
	File*		pf = NULL;
    int			ret;


    THR_ENTER_LIB();

    if (!thr_init_done) {
        thr_init();
    }
    assert(thr_curr);

	/* validate our input */
    if (!File::IsValid(fd)) {
        thr_report_error("bad fd %d, cannot bind to t%u\n",
            fd, thr_curr->id.tid);
		RETURN( THR_EBADFD );
    }

	/* check that we were given a place for the return value */
	if( ptid == NULL )
	{
		thr_report_error("cannot return bound fd's thread_t\n");
		RETURN( THR_ESRCH );
	}

	/* determine whether file descriptor is already bound */
	pf = File::Find( fd );
	if( (pf != NULL) && (pf->tid != thr_curr->id.tid) )
	{
        thr_report_error("cannot rebind fd %d, already bound to t%u\n",
            fd, pf->tid);
        RETURN( THR_EBADFD );
	}

	// obtain and initialize a new File
	pf = File::Alloc( fd, thr_curr->id.tid, special, will_block, desc, ptid );

	// add the new item to our local waitset
	thr_curr->Bind( *ptid );

    CHECK_REALLY_RESCHEDULE();
    DO_TRACE_0(("msg_bind(%u):", fd));

    ret = thr_err( THR_OKAY );

finish:
    THR_LEAVE_LIB();
    return ret;
}







/************************************************************************
 * int msg_bind_socket(PDSOCKET sd, unsigned special,
 *              int (*will_block)(void *), void* desc, thread_t* ptid)
 *
 * Bind socket descriptor `sd' to the calling thread so that all _input_
 * coming through this socket is available as a message to the thread.
 *
 * The `special' flag indicates how `sd' should be treated.  If not set,
 * bytes appearing on this socket will be copied to the receiver's buffers.
 * If `special' is set, the receiver will get an indication of the
 * socket's readiness for reading but no bytes will be dequeued.
 *
 * If a non-null will_block function pointer is given, libthread assumes
 * that the data on the stream `sd' can be buffered, and uses 
 * `will_block(desc)' to determine if no buffered data is
 * present for the descriptor in question.
 *
 * The thread_t assigned to the given socket descriptor is returned
 * in *ptid.  It is an error to specify NULL for ptid.
************************************************************************/

int
msg_bind_socket(PDSOCKET s,
	unsigned special,
	int (*will_block)(void *),
	void* desc,
	thread_t* psid)
{
	Socket*		ps = NULL;
    int			ret;


    THR_ENTER_LIB();

    if (!thr_init_done) {
        thr_init();
    }
    assert(thr_curr);

	/* validate input */
    if (!Socket::IsValid(s)) {
        thr_report_error("bad SOCKET %u, cannot bind to t%u\n",
            s, thr_curr->id.tid);
		RETURN( THR_EBADSOCK );
    }

	if( psid == NULL )
	{
		thr_report_error("cannot return bound SOCKET's tid\n");
		RETURN( THR_ESRCH );
	}

	/* determine whether socket is already bound */
	ps = Socket::Find( s );
	if( (ps != NULL) && (ps->tid != thr_curr->id.tid) )
	{
        thr_report_error("cannot rebind SOCKET %u, already bound to t%u\n",
            s, ps->tid);
        RETURN( THR_EBADSOCK );
	}

	/* obtain and initialize a new item for this socket */
	ps = Socket::Alloc( s, thr_curr->id.tid, special, will_block, desc, psid );

	// add the new item to our local waitset
	thr_curr->Bind( *psid );

    CHECK_REALLY_RESCHEDULE();
    DO_TRACE_0(("msg_bind_socket(%u):", s));

    ret = thr_err( THR_OKAY );

finish:
    THR_LEAVE_LIB();
    return ret;
}



#if defined(i386_unknown_nt4_0)
/************************************************************************
 * int msg_bind_wmsg(thread_t* tid)
 *
 * Bind the calling thread so that all messages from the Windows message
 * queue are available as messages to the thread.
************************************************************************/

int
msg_bind_wmsg(thread_t* ptid)
{
    int ret = THR_OKAY;
	WindowsMsgQ* pwmsg;
	thread_t wmqtid;


    THR_ENTER_LIB();

    if (!thr_init_done) {
        thr_init();
    }
    assert(thr_curr);

	// validate input
	if( ptid == NULL )
	{
		thr_report_error("cannot return bound tid\n");
		RETURN( THR_ESRCH );
	}

	// determine whether the Item is already bound
	pwmsg = WindowsMsgQ::Find( &wmqtid );
	if( (pwmsg != NULL) && (pwmsg->tid != thr_curr->id.tid) )
	{
        thr_report_error("cannot rebind Windows message queue, already bound to t%u\n",
            pwmsg->tid);
        RETURN( THR_EPARAM );
	}

	if( pwmsg == NULL )
	{
		// obtain and initialize a new item for the Windows message queue
		pwmsg = WindowsMsgQ::Alloc( thr_curr->id.tid, ptid );

		// add the new item to our local waitset
		thr_curr->Bind( *ptid );
	}
	else
	{
		// the Item was already bound to this thread, so just return the tid
		*ptid = wmqtid;
	}

    CHECK_REALLY_RESCHEDULE();
    DO_TRACE_0(("msg_bind_wmsg():"));

    ret = thr_err( THR_OKAY );

finish:
    THR_LEAVE_LIB();
    return ret;
}

#endif // defined(i386_unknown_nt4_0)



/************************************************************************
 * int msg_unbind(thread_t tid)
 *
 * Removes the binding of the bindable represented by 'tid' from the
 * current thread.  The bindable will no longer be inspected for input.
************************************************************************/

int
msg_unbind(thread_t tid)
{
    int ret;

    THR_ENTER_LIB();

    if (!thr_init_done) {
        thr_init();
    }
    assert(thr_curr);

	if( !thr_tab->IsValid( tid ) )
	{
        thr_report_error("bad tid %u, cannot unbind from t%u\n",
            tid, thr_curr->id.tid);
        RETURN(THR_ESRCH);
	}

	// note - we know that GetItem is safe due to the previous IsValid check
	if( !(thr_tab->GetItem(tid)->IsBindable()) )
	{
        thr_report_error("cannot unbind non-bindable tid %u\n", tid);
        RETURN(THR_ESRCH);
    }

    if (!thr_curr->i_own(tid)) {
        thr_report_error("cannot unbind non-owned bindable %u\n", tid);
        RETURN(THR_ESRCH);
    }

	// do the actual unbinding
	thr_curr->Unbind(tid);

	// remove the item from the Item table
	thr_tab->Dealloc( tid );

    CHECK_REALLY_RESCHEDULE();
    DO_TRACE_0(("msg_unbind(%u):", tid));
    ret = thr_err(THR_OKAY);

finish:
    THR_LEAVE_LIB();
    return ret;
}





/************************************************************************
 * void msg_dump_state(void)
 *
 * dump the current state of the messages in libthread to `stderr'.
 * the state of the pending message queues (tag, sender, size) of
 * each thread is printed.
************************************************************************/

#if defined(THREADLIB_DEBUG)
void
msg_dump_state(void)
{
	cerr << "msg state {" << endl;

	// dump global waitset
	cerr << "global waitset: " << thr_all_waitset;

	// dump the item table
	cerr << "item table: " << *thr_tab;

	cerr << '}' << endl;
}
#endif // defined(THREADLIB_DEBUG)



/************************************************************************
 * utility functions
************************************************************************/

/************************************************************************
 * void msg_copy(void* to_buf, unsigned* to_sizep, void* from_buf,
 *               unsigned from_size)
 *
 * copy the minimum of `*to_sizep' and `from_size' bytes from `from_buf'
 * to `to_buf'. print a warning if the destination buffer size is less
 * than the source buffer size. `*to_sizep' is set to the number of
 * bytes copied.
 *
 * this function is used by both message senders and receivers to copy
 * buffers.
************************************************************************/

static void
msg_copy(void* to_buf, unsigned* to_sizep, void* from_buf, unsigned from_size)
{
    if ((to_buf == 0) || (to_sizep == 0)) {
        return;
    }

    if (*to_sizep < from_size) {
        thr_report_error("dst-size `%u' < src-size `%u'\n", *to_sizep,
            from_size);
    }

    *to_sizep = min(*to_sizep, from_size);
    (void) memcpy(to_buf, from_buf, *to_sizep);

    return;
}





/************************************************************************
 * void queue_msg(tag_t tag, void* buf, unsigned size, Thread t)
 *
 * queue the message `tag,buf,size' with thread `t'. allocate a new
 * message buffer and add it to the target thread's queue.
************************************************************************/

static void
queue_msg(tag_t tag, void* buf, unsigned size, Thread t)
{
    ThrMessage* msg;


    if (!(msg = (ThrMessage*) malloc(sizeof(ThrMessage)))) {
        perror("malloc");
        abort();
    }

    if (size > 0) {
        if (!(msg->buf = (void *) malloc(size))) {
            perror("malloc");
            abort();
        }
        (void) memcpy(msg->buf, buf, size);
    }
    else {
        msg->buf = 0;
    }

    msg->size   = size;
    msg->tag    = tag;
    msg->sender = thr_curr->id.tid;

    if (!t->msg.mq) {
        NEWQ(t->msg.mq);
    }
    ENQUEUE(t->msg.mq, msg);

    return;
}






/************************************************************************
 * ThrMessage* find_match(thread_t tid, tag_t tag, unsigned do_poll)
 *
 * scan through the message queue of the current thread and check to
 * see if 'tid' and 'tag' matches any of the messages in the queue.
 *
 * on a match, dequeue the message and return a pointer to it, else
 * return null.
 *
 * if the boolean flag `do_poll' is set, then no state changes are made
 * and the message is not removed from the message queue.
************************************************************************/

static ThrMessage*
find_match(thread_t tid, tag_t tag, unsigned do_poll)
{
    qscan_t qs;
    ThrMessage* msg;


    assert(thr_curr);

    if (!thr_curr->msg.mq) {
        return 0;
    }

    qs = FIRST(thr_curr->msg.mq);
    while (!EOSCAN(thr_curr->msg.mq, qs)) {
        msg = (ThrMessage*) GET(thr_curr->msg.mq, qs);
		if (msg_match(msg->sender, msg->tag, tid, tag) ) {
            if (!do_poll) {
                msg = (ThrMessage*) REMOVE(thr_curr->msg.mq, qs);
            }
            return msg;
        }
        qs = NEXT(thr_curr->msg.mq, qs);
    }

    return 0;
}





/************************************************************************
 * bool got_sigmsg(tag_t* tagp, unsigned* nitemsp, int* senderp,
 *                     unsigned do_poll)
 *
 * disable signals and search through the set of signals for those
 * bound to the current thread and for any pending messages on one of
 * these signals. the first ready signal is processed.
 *
 * `*tagp' should be compatible with signal messages. `*senderp' is set
 * to the ready signal (if any) and `*nitemsp' is set to the number of
 * signals received so far.
 *
 * if the boolean flag `do_poll' is set, no state changes are made.
 * the function returns `1' if signal messages are found and `0' otherwise.
************************************************************************/

static bool
got_sigmsg(tag_t* tagp, unsigned* nitemsp, int* senderp,
           unsigned do_poll)
{
#if SIG_READY
    unsigned i;


    assert(thr_curr);
    assert(is_sig_tag(*tagp));

    thr_sig_enter_cs();

    if (thr_curr->sig.nsigs == 0) {
        thr_sig_leave_cs();
        return false;
    }

    for (i = 0; i < thr_sig_map_size; i++) {
        if (i_own_sig(i) && (thr_sig_map[i].count > 0)) {
            *tagp    = MSG_TAG_SIG;
            *senderp = ndx2sig(i);

            if (!do_poll) {
                if (nitemsp) {
                    *nitemsp = thr_sig_map[i].count;
                }

                thr_curr->sig.nsigs -= thr_sig_map[i].count;
                thr_sig_map[i].count = 0;
            }

            thr_sig_leave_cs();
            return true;
        }
    }

    thr_report_error("inconsistency between nsigs and counts\n");
    abort();

    /* NOTREACHED */
    thr_sig_leave_cs();
#endif // SIG_READY
    return true;
}



/************************************************************************
 * bool got_qmsg(thread_t* tidp, tag_t* tagp, void* buf,
 *					 unsigned* sizep, int* senderp, unsigned do_poll)
 *
 * check the message queue of the current thread for waiting messages.
 * if a message matching '*tidp' and '*tagp' is found, dequeue the message
 * and copy it into `*buf' and set the size `*sizep' and the sender of the
 * message as `*tidp'.
 *
 * if the boolean flag `do_poll' is set, no state changes are made.
 *
 * if a message was found and dequeued, the return value is a boolean
 * true, else it is false.
************************************************************************/

static bool
got_qmsg(thread_t* tidp, tag_t* tagp, void* buf, unsigned* sizep,
		 unsigned do_poll)
{
    ThrMessage* msg;

    assert(thr_curr);
	assert(tagp);
	assert(tidp);
	assert(SpecifiesThread(*tidp, *tagp));

    msg = find_match(*tidp, *tagp, do_poll);
    if (!msg) {
        return false;
    }

	*tagp		= msg->tag;
	*tidp		= msg->sender;

    if (!do_poll) {
        msg_copy(buf, sizep, msg->buf, msg->size);

        if (msg->buf) {
            (void) free(msg->buf);
        }
        (void) free(msg);
    }

    return true;
}









/************************************************************************
 * bool got_filemsg(thread_t* tidp, tag_t* tagp, void* buf,
 *						unsigned* sizep, unsigned do_poll)
 *
 * initially, check all bound and buffered files (via
 * their "will_block" functions) to see if any messages are ready on
 * them. if so, return the first such descriptor.
 *
 * check the files bound to the current thread for waiting
 * messages. if any descriptors are ready for reading, either dequeue
 * the message (for non-special files) or return an indication of the
 * readiness.
 *
 * for non-special files, the waiting message is copied into `*buf'
 * and `*sizep' is set to the number of bytes read in.
 *
 * for both special and non-special files, '*tidp' is set to the file
 * that is ready and '*tagp' to 'MSG_TAG_FILE'.
 *
 * the return value is true if a file message was dequeued (or found
 * ready) and false if no file was ready for reading.
 *
 * if the boolean flag `do_poll' is set, no state changes are made and
 * no bytes are dequeued from the file (even for non-special files).
************************************************************************/

static bool
got_filemsg(thread_t* tidp, tag_t* tagp, void* buf, unsigned* sizep,
		  bool do_poll)
{
	WaitSet	tmp_waitset;

    assert(thr_curr);
    assert(tagp);
    assert(tidp);
	assert(SpecifiesFile(*tidp, *tagp));

	
	// check whether a specific source was specified
	if( !is_unspec_tid(*tidp) )
	{
		// a specific source was specified
		if( thr_curr->i_own( *tidp ) &&
			thr_tab->GetIOBindable( *tidp )->IsBufferReady( sizep, do_poll ) )
		{
			*tagp = MSG_TAG_FILE;
			return true;
		}
    }
    else
	{
		thread_t fid;

		// no specific source was specified - try all of our bound files
		if( thr_curr->fd.last_file == THR_TID_UNSPEC )
		{
			// look for the first bound file
			fid = thr_tab->FindNextOwned( THR_TID_UNSPEC, item_t_file );
			if( fid == THR_TID_UNSPEC )
			{
				// there are no bound files that we own
				return false;
			}
			else
			{
				// start our last_known bound file search at the file we found
				thr_curr->fd.last_file = fid;
			}
		}
		else
		{
			// start at our last known bound file
			fid = thr_curr->fd.last_file;
		}
		assert( thr_tab->IsFile( fid ) );

		do {
			assert( thr_curr->i_own( fid ) );

			if( thr_tab->GetIOBindable( fid )->IsBufferReady( sizep, do_poll ) )
			{
				*tidp = fid;
				*tagp = MSG_TAG_FILE;
				thr_curr->fd.last_file = thr_tab->FindNextOwned( fid, item_t_file );
				return true;
			}

			// move to next file
			fid = thr_tab->FindNextOwned( fid, item_t_file );
		} while (fid != thr_curr->fd.last_file );
    }
    
	// no buffered messages were available
	// check whether any file descriptors are ready to read
    if (!is_unspec_tid(*tidp)) {
		// consider the specific source
		tmp_waitset.Add( *tidp );
    }
    else {
		// consider all bound files
		add_all_bytype( tmp_waitset, item_t_file );
	}

    if (thr_is_ready_io(tmp_waitset, false, thr_curr->fd.last_file, tidp, tagp)) {
        assert(thr_curr->i_own(*tidp));
		assert(*tagp == MSG_TAG_FILE);

        if (!do_poll) {

			thr_tab->GetBindable( *tidp )->HandleMessage( buf, sizep );
        }
		*tagp = MSG_TAG_FILE;

		// advance our "last file" indicator
		thr_curr->fd.last_file = thr_tab->FindNextOwned( *tidp, item_t_file );

        return true;
    }

    return false;
}





/************************************************************************
 * bool got_socketmsg(thread_t* tidp, tag_t* tagp, void* buf,
 *						unsigned* sizep, bool do_poll)
 *
 * initially, check all bound and buffered sockets (via
 * their "will_block" functions) to see if any messages are ready on
 * them. if so, return the first such descriptor.
 *
 * check the sockets bound to the current thread for waiting
 * messages. if any descriptors are ready for reading, either dequeue
 * the message (for non-special files) or return an indication of the
 * readiness.
 *
 * for non-special sockets, the waiting message is copied into `*buf'
 * and `*sizep' is set to the number of bytes read in.
 *
 * for both special and non-special sockets, '*tidp' is set to the socket
 * that is ready and '*tagp' to 'MSG_TAG_SOCKET'.
 *
 * the return value is true if a socket message was dequeued (or found
 * ready) and false if no socket was ready for reading.
 *
 * if the boolean flag `do_poll' is set, no state changes are made and
 * no bytes are dequeued from the socket (even for non-special files).
************************************************************************/

static bool
got_socketmsg(thread_t* tidp, tag_t* tagp, void* buf, unsigned* sizep,
		  bool do_poll)
{
	WaitSet	tmp_waitset;

    assert(thr_curr);
    assert(tagp);
    assert(tidp);
	assert(SpecifiesSocket(*tidp, *tagp));

	
	// check whether a specific source was specified
	if( !is_unspec_tid(*tidp) )
	{
		// a specific source was specified
		if( thr_curr->i_own( *tidp ) &&
			thr_tab->GetIOBindable( *tidp )->IsBufferReady( sizep, do_poll ) )
		{
			*tagp = MSG_TAG_SOCKET;
			return true;
		}
    }
    else
	{
		thread_t sockid;

		// no specific source was specified - try all of our bound sockets
		if( thr_curr->fd.last_socket == THR_TID_UNSPEC )
		{
			// look for the first bound socket
			sockid = thr_tab->FindNextOwned( THR_TID_UNSPEC, item_t_socket );
			if( sockid == THR_TID_UNSPEC )
			{
				// there are no bound sockets that we own
				return false;
			}
			else
			{
				// start our last_known bound socket search at the socket we found
				thr_curr->fd.last_socket = sockid;
			}
		}
		else
		{
			// start at our last known bound socket
			sockid = thr_curr->fd.last_socket;
		}
		assert( thr_tab->IsSocket( sockid ) );

		do {
			assert( thr_curr->i_own( sockid ) );

			if( thr_tab->GetIOBindable( sockid )->IsBufferReady( sizep, do_poll ) )
			{
				*tidp = sockid;
				*tagp = MSG_TAG_SOCKET;
				thr_curr->fd.last_socket = thr_tab->FindNextOwned( sockid, item_t_socket );
				return true;
			}

			// move to next socket
			sockid = thr_tab->FindNextOwned( sockid, item_t_socket );
		} while (sockid != thr_curr->fd.last_socket );
    }
    
	// no buffered messages were available
	// check whether any socket descriptors are ready to read
    if (!is_unspec_tid(*tidp)) {
		// consider the specific source
		tmp_waitset.Add( *tidp );
    }
    else {
		// consider all bound sockets
		add_all_bytype( tmp_waitset, item_t_socket );
	}

    if (thr_is_ready_io(tmp_waitset, false, thr_curr->fd.last_socket, tidp, tagp)) {
        assert(thr_curr->i_own(*tidp));
		assert(*tagp == MSG_TAG_SOCKET);

        if (!do_poll) {

			thr_tab->GetBindable( *tidp )->HandleMessage( buf, sizep );
        }
		*tagp = MSG_TAG_SOCKET;

		// advance our "last socket" indicator
		thr_curr->fd.last_socket = thr_tab->FindNextOwned( *tidp, item_t_socket );

        return true;
    }

    return false;
}







/************************************************************************
 * int chk_msg_pref(const char* func, thread_t* tidp, tag_t* tagp,
 *             void* buf, unsigned* nitemsp, bool do_poll,
 *             bool do_block, unsigned io_first)
 *
 * the common path for both a `msg_recv,' `msg_poll,' and 
 * 'msg_poll_preference.' during a receive, `tagp, buf, nitemsp' are 
 * valid and `do_poll, do_block' are false. during a poll, `tagp' is
 * valid, `buf, nitemsp' are null and `do_poll' is true.  uses the 
 * value of `io_first' to determine whether or not to check for a 
 * msg from a file or socket before a thread msg (only used by
 * msg_poll_preference).
************************************************************************/

static int
chk_msg_pref(const char* func, thread_t* tidp, tag_t* tagp,
			 void* buf, unsigned* nitemsp,
			 bool do_poll, bool do_block, unsigned io_first)
{
    WaitSet merged_wset;


    if (!thr_init_done) {
        thr_report_error("*** thread library not initialized ***\n");
        return thr_err(THR_ERR);
    }
    assert(thr_curr);

	// check if we've already seen a message of the desired type
    DO_TRACE_0(("%s: checking for waiting messages", func));

#if SIG_READY
    if (is_sig_tag(*tagp)) {
        if (got_sigmsg(tagp, nitemsp, &ret, do_poll)) {
            CHECK_RESCHEDULE();
            DO_TRACE_0(("%s: found signal `%d' messages", func, ret));
            return ret;
        }
    }
#endif // SIG_READY

    if(io_first){

		if( SpecifiesFile(*tidp, *tagp) )
		{
            if (got_filemsg(tidp, tagp, buf, nitemsp, do_poll)) {
                CHECK_RESCHEDULE();
                DO_TRACE_0(("%s: found file message on `%d'", func, *tidp));
                return thr_err(THR_OKAY);
            }
		}

		if( SpecifiesSocket(*tidp, *tagp) )
		{
            if (got_socketmsg(tidp, tagp, buf, nitemsp, do_poll)) {
                CHECK_RESCHEDULE();
                DO_TRACE_0(("%s: found socket message on `%d'", func, *tidp));
                return thr_err(THR_OKAY);
            }
		}

		if( SpecifiesThread(*tidp, *tagp) )
		{
            if (got_qmsg(tidp, tagp, buf, nitemsp, do_poll)) {
                CHECK_RESCHEDULE();
                DO_TRACE_0(("%s: found queue message from `%d'", func, *tidp));
                return thr_err(THR_OKAY);
            }
        }
    }
    else{
		if( SpecifiesThread(*tidp, *tagp) )
		{
            if (got_qmsg(tidp, tagp, buf, nitemsp, do_poll)) {
                CHECK_RESCHEDULE();
                DO_TRACE_0(("%s: found queue message from `%d'", func, *tidp));
                return thr_err(THR_OKAY);
            }
        }

		if( SpecifiesFile(*tidp, *tagp) )
		{
            if (got_filemsg(tidp, tagp, buf, nitemsp, do_poll)) {
                CHECK_RESCHEDULE();
                DO_TRACE_0(("%s: found file message on `%d'", func, *tidp));
                return thr_err(THR_OKAY);
            }
		}

		if( SpecifiesSocket(*tidp, *tagp) )
		{
            if (got_socketmsg(tidp, tagp, buf, nitemsp, do_poll)) {
                CHECK_RESCHEDULE();
                DO_TRACE_0(("%s: found socket message on `%d'", func, *tidp));
                return thr_err(THR_OKAY);
            }
		}
    }

	// we hadn't already seen a message of the desired type
	// check if we have to merge our waitset into the
	// global waitset so that our bindables are checked
    if( is_io_tag(*tagp) )
	{
		// did caller specify a specific source?
		if( !is_unspec_tid(*tidp) )
		{
			// caller specified a particular source, named by *tidp
			if( !thr_curr->i_own( *tidp ) )
			{
				// TODO - is this correct in cases where tag is any?
				// we do not own the source if it is from a thread
                thr_report_error("cannot receive from unbound item `%d'\n",
                    (int) *tidp);
                return thr_err(THR_ESRCH);
			}

			// add just this bindable to the global waitset
			merged_wset.Add( *tidp );
            DO_TRACE_0(("%s: merging tid `%u' into global set", func, *tidp));
		}
		else
		{
			// caller didn't specify a particular source
			// did they specify a particular source type?
			if( is_unspec_tag( *tagp ) )
			{
				// caller didn't specify a source type either
				add_all( merged_wset );
				DO_TRACE_0(("%s: merging all bound items into global set", func));
			}
			else if( is_file_tag( *tagp ) )
			{
				// caller specified file as the source type
				add_all_bytype( merged_wset, item_t_file );
				DO_TRACE_0(("%s: merging all bound files into global set", func));
			}
			else if( is_socket_tag( *tagp ) )
			{
				// caller specified socket as the source type
				add_all_bytype( merged_wset, item_t_socket );
				DO_TRACE_0(("%s: merging all bound sockets into global set", func));
			}
			else
			{
				// should never get here
				thr_report_error( "*** unrecognized tag type requested ***\n" );
				abort();
			}
		}

		// add specified bindables to the global waitset
		thr_all_waitset.Add( merged_wset );
   }

    if (do_poll && !do_block) {
        CHECK_RESCHEDULE();
        DO_TRACE_0(("%s: no messages found", func));
        return thr_err(THR_ENOMSG);
    }

	thr_curr->msg.tidp		= tidp;
	thr_curr->msg.tagp		= tagp;
    thr_curr->msg.buf       = buf;
    thr_curr->msg.nitemsp   = nitemsp;
    thr_curr->flags.blocked = 1;
    thr_curr->flags.polling = do_poll;

    DO_TRACE_0(("%s: no pending messages, blocking", func));
	fflush( stderr );

    thr_reschedule();

    DO_TRACE_0(("%s: unblocked by sender `%d'", func, *tidp ));

	// remove any bindables we merged into the global set
	if( !merged_wset.IsEmpty() )
	{
		DO_TRACE_0(("%s: deleting merged bindables from global set", func));
		thr_all_waitset.Remove( merged_wset );
	}

    return thr_err(THR_OKAY);
}





static
void
add_all_bytype( WaitSet& wset, item_t itype )
{
	ThrTableIterator	iter( *thr_tab );
	thread_t tid;

	
	while( iter.Next( itype, &tid ) )
	{
		if( thr_curr->i_own(tid) )
		{
			wset.Add( tid );
		}
	}
}


static
void
add_all( WaitSet& wset )
{
	ThrTableIterator	iter( *thr_tab );
	thread_t tid;

	
	while( iter.Next( &tid ) )
	{
		if( thr_curr->i_own(tid) )
		{
			wset.Add( tid );
		}
	}
}


